#include <pvOptFuncs.h>

#include <fstream>
#include <sstream>
#include <string>
#include <iostream>
#include <vector>

#include <QDebug>
#include <QFile>
#include <QFileInfo>
#include <QByteArray>
#include <QString>

using namespace std;

struct vertex {
	float x;
	float y;
	float z;
};

vertex operator ^ (const vertex &, const vertex &);
vertex operator - (const vertex &, const vertex &);
std::vector<float> findBounds(std::vector<std::vector<vertex>>);
std::vector<float> findLocation(std::vector<std::vector<vertex>>,int);
float vertexNorm(vertex);
std::vector<std::vector<vertex>> binarySTLImport(const string &);
std::vector<std::vector<vertex>> asciiSTLImport(std::string);
std::vector<std::vector<vertex>> STLImport(const QString &);
void unit(vertex &);

bool isSTL(std::string);
int getStlFileFormat(const QString &);
//bool isBinarySTL(const string &);

std::vector<std::vector<vertex>> binarySTLImport(const string & fileName) {

    std::vector<std::vector<vertex>> triangles;

	int nVertex = 0; // Number of vertices read.
	int nFacet = 0;  // Number of facets read.

	// Open the file for reading using an input fstream.
    ifstream ifs(fileName, ifstream::binary);

	// Get pointer to the associated buffer object.
	// rdbuf returns a streambuf object associated with the
	// input fstream object ifs.

	filebuf* pbuf = ifs.rdbuf();

	// Calculate the file's size.

	auto size = pbuf->pubseekoff(0, ifs.end);

	// Set the position pointer to the beginning of the file.

	pbuf->pubseekpos(0);

	// Allocate memory to contain file data.

	char* buffer = new char[(size_t)size];

	// Get file data. sgetn grabs all the characters from the streambuf 
	// object 'pbuf'. The return value of sgetn is the number of characters
	// obtained - ordinarily, this value should be checked for equality
	// against the number of characters requested.

	pbuf->sgetn(buffer, size);

	char * bufptr = buffer;

	bufptr += 80;  // Skip past the header.
	bufptr += 4;   // Skip past the number of triangles.

	vertex normal;
	vertex v1, v2, v3;

	while (bufptr < buffer + size) {

		normal.x = *(float *)(bufptr);
		normal.y = *(float *)(bufptr + 4);
		normal.z = *(float *)(bufptr + 8);
		bufptr += 12;

		v1.x = *(float *)(bufptr);
		v1.y = *(float *)(bufptr + 4);
		v1.z = *(float *)(bufptr + 8);
		bufptr += 12;

		v2.x = *(float *)(bufptr);
		v2.y = *(float *)(bufptr + 4);
		v2.z = *(float *)(bufptr + 8);
		bufptr += 12;

		v3.x = *(float *)(bufptr);
		v3.y = *(float *)(bufptr + 4);
		v3.z = *(float *)(bufptr + 8);
		bufptr += 12;

		const float eps = (float) 1.0e-9;

		// If the normal in the STL file is blank, then create a proper normal.

		if (abs(normal.x) < eps && abs(normal.y) < eps && abs(normal.z) < eps) {
			vertex u, v;
			u = v2 - v1;
			v = v3 - v1;
			normal = u ^ v;
			unit(normal);
		}

		nFacet++;
		nVertex += 3;

		//void STLAddFacet(vertex, vertex, vertex, vertex);
		//STLAddFacet(v1, v2, v3, normal);

        std::vector<vertex> thisTriangle{ v1, v2, v3, normal };
        triangles.push_back(thisTriangle);

		bufptr += 2;
	}


	ifs.close();

	delete[] buffer;
    return triangles;
}

std::vector<std::vector<vertex>> asciiSTLImport(std::string fileName)
{
  std::vector<std::vector<vertex>> triangles;

  std::ifstream asciiSTL(fileName.c_str(), std::ios::in);
  std::string line;

  while (std::getline(asciiSTL, line))
  {

    vertex normal;
    vertex v1, v2, v3;

    //facet loop
	std::vector<std::string> split = pvOptFuncs::splitString(line, " ");
	auto findFacet = std::find(split.begin(), split.end(), "facet");
    int splitNum = findFacet - split.begin();
	if (findFacet != split.end())
	{  
      normal.x = std::stof(split[splitNum+2]);
	  normal.y = std::stof(split[splitNum+3]);
	  normal.z = std::stof(split[splitNum+4]);
	  int vertexSwitch = 0;
	  do
	  {
	    std::getline(asciiSTL, line);
		split = pvOptFuncs::splitString(line, " ");
		auto findVertex = std::find(split.begin(), split.end(), "vertex");
		splitNum = findVertex - split.begin();
		if (findVertex != split.end())
		{
		  if (vertexSwitch == 0)
		  {
		    v1.x = std::stof(split[splitNum+1]);
	        v1.y = std::stof(split[splitNum+2]);
	        v1.z = std::stof(split[splitNum+3]);
		  }
		  else if (vertexSwitch == 1)
		  {
			v2.x = std::stof(split[splitNum+1]);
	        v2.y = std::stof(split[splitNum+2]);
	        v2.z = std::stof(split[splitNum+3]);
		  }
		  else if (vertexSwitch == 2)
		  {
			v3.x = std::stof(split[splitNum+1]);
	        v3.y = std::stof(split[splitNum+2]);
	        v3.z = std::stof(split[splitNum+3]);
		  }
		  vertexSwitch++;
		}
	  } while (vertexSwitch < 3);
    
      std::vector<vertex> thisTriangle{ v1, v2, v3, normal };
      triangles.push_back(thisTriangle);
	}

  }

  asciiSTL.close();

  return triangles;
}

std::vector<std::vector<vertex>> STLImport(const QString & filePath)
{
  int i = getStlFileFormat(filePath);
  if (i == 0)
  {
    return binarySTLImport(filePath.toStdString().c_str());
  }
  else if (i == 1)
  {
    return asciiSTLImport(filePath.toStdString());
  }
  qDebug() << "Error opening stl file";
  std::vector<std::vector<vertex>> err;
  return err;
}

std::vector<float> findBounds(std::vector<std::vector<vertex>> triangles)
{
  std::vector<float> bounds{ 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 }; //xMin xMax yMin yMax zMin zMax
  for (int i = 0; i < triangles.size(); i++)
  {
    std::vector<vertex> triangleVertices = triangles[i];
    for (int j = 0; j < triangleVertices.size() - 1; j++)
    {
      vertex vertice = triangleVertices[j];
      if (i == 0 && j== 0)
      {
        bounds[0] = vertice.x;
        bounds[1] = vertice.x;
        bounds[2] = vertice.y;
        bounds[3] = vertice.y;
        bounds[4] = vertice.z;
        bounds[5] = vertice.z;
      }

      if (vertice.x < bounds.at(0))
      {
        bounds[0] = vertice.x;
      }
      else if (vertice.x > bounds.at(1))
      {
        bounds[1] = vertice.x;
      }

      if (vertice.y < bounds.at(2))
      {
        bounds[2] = vertice.y;
      }
      else if (vertice.y > bounds.at(3))
      {
        bounds[3] = vertice.y;
      }

      if (vertice.z < bounds.at(4))
      {
        bounds[4] = vertice.z;
      }
      else if (vertice.z > bounds.at(5))
      {
        bounds[5] = vertice.z;
      }
    }
  }
  return bounds;
}

std::vector<float> findLocation(std::vector<std::vector<vertex>> triangles, int location)
{
  //location codes? 0 outside, 1 inside

  //find vertice of triangle with highest x,y,z
  vertex furthestVertice;
  std::vector<vertex> furthestTriangle;
  std::vector<float> loc{ 0.0, 0.0, 0.0 };
  furthestVertice.x = 0;
  furthestVertice.y = 0;
  furthestVertice.z = 0;

  for (int i = 0; i<triangles.size(); i++)
  {
    std::vector<vertex> triangle = triangles[i];
    for (int j = 0; j<triangle.size()-1; j++)
    {
      vertex vertice = triangle[j];
      if (vertexNorm(vertice) > vertexNorm(furthestVertice))
      {
        furthestVertice = vertice;
        furthestTriangle = triangle;
      }
    }
  }
  
  if (location == 0) //outside
  {
    //add
    float mag = vertexNorm(furthestVertice)*0.05; //5% out of mesh?
    loc[0] = furthestVertice.x*(1+mag/abs(furthestVertice.x));
    loc[1] = furthestVertice.y*(1+mag/abs(furthestVertice.y));
    loc[2] = furthestVertice.z*(1+mag/abs(furthestVertice.z));
    return loc;
  }
  else if (location == 1) //inside
  {
    //subtract along normal?
    float mag = vertexNorm(furthestVertice)*0.01; // 1% into mesh?
    loc[0] = furthestVertice.x-(mag*furthestTriangle.back().x);
    loc[1] = furthestVertice.y-(mag*furthestTriangle.back().y);
    loc[2] = furthestVertice.z-(mag*furthestTriangle.back().z);
    return loc;
  }
  return std::vector<float>{};
}

float vertexNorm(vertex v)
{
  return sqrt(pow(v.x,2)+pow(v.y,2)+pow(v.z,2));
}

vertex operator ^ (const vertex & a, const vertex & b) {
	// Cross product.
	vertex result;
	result.x = a.y * b.z - a.z * b.y;
	result.y = a.z * b.x - a.x * b.z;
	result.z = a.x * b.y - a.y * b.x;
	return(result);
}

vertex operator - (const vertex & a, const vertex & b) {
	// Subtraction.
	vertex result;
	result.x = a.x - b.x;
	result.y = a.y - b.y;
	result.z = a.z - b.z;
	return(result);
}

void unit(vertex & v) {
	// Normalize a vector.
	float vmod = pow(v.x, 2) + pow(v.y, 2) + pow(v.z, 2);
	vmod = sqrt(vmod);

	if (vmod > (float)1.0e-9) {
		v.x /= vmod;
		v.y /= vmod;
		v.z /= vmod;
	}
}

/*bool isBinarySTL(const string & filePath) {

  ifstream ifs(filePath, ifstream::binary);
	// Get pointer to the associated buffer object.
  // rdbuf returns a streambuf object associated with the
  // input fstream object ifs.
  filebuf* pbuf = ifs.rdbuf();
  // Calculate the file's size.
  auto size = pbuf->pubseekoff(0, ifs.end);
  char* buffer = new char[(size_t)size];
	// Determine if a file is binary STL.

	bool bbinary = true;
	size_t spnsz, spnsz0;

	// Look for the first non-space character.

	spnsz = strspn(buffer, " ");

	char ctstr[6];  // Enough space to hold "solid\0" and "facet\0".

	// Copy the first five characters from the location of the first non-space
	// character to ctstr.

	strncpy(ctstr, &buffer[spnsz], 5);

	ctstr[5] = '\0';
	char csolid[] = "solid\0";

	// If this is an ASCII STL file, then the first string should be "solid".

	if (!strcmp(ctstr, csolid)) {
		// This file might be binary or text. To be certain we need to do a further test.
		// Read past the next new line. If this is a text file, there should be a newline.
		// The next token should be 'facet'.

		spnsz0 = 5 + spnsz;

		char * pch = strchr(&buffer[spnsz0], '\n');  // Look for the first instance of '\n'.
		// If pch is NULL then this is a binary STL file.
		if (pch) {
			pch++;

			spnsz = strspn(pch, " "); // Look for the first instance not of ' '.
			spnsz0 = spnsz;

			spnsz = strcspn(pch + spnsz0, " "); // Look for the first instance of ' '.

			if (spnsz == 5) {
				// Check for 'facet'.
				strncpy(ctstr, pch + spnsz0, 5);
				ctstr[5] = '\0';

				char cfacet[] = "facet\0";
				if (!strcmp(ctstr, cfacet)) {
					// This file is beyond reasonable doubt ASCII STL.
					bbinary = false;
				}
			}
		}
	}

	return(bbinary);


}*/

int getStlFileFormat(const QString &path)
{
    QFile file(path);
    if (!file.open(QIODevice::ReadOnly))
    {
        qDebug("\n\tUnable to open \"%s\"", qPrintable(path));
        return -1;
    }

    QFileInfo fileInfo(path);
    size_t fileSize = fileInfo.size();

    // Look for text "solid " in first 6 bytes, indicating the possibility that this is an ASCII STL format.

    if (fileSize < 15)
    {
        // "solid " and "endsolid " markers for an ASCII file
        qDebug("\n\tThe STL file is not long enough (%u bytes).", uint(fileSize));
        return -1;
    }

    // binary files should never start with "solid ", but
    // just in case, check for ASCII, and if not valid then
    // check for binary...
 
    QByteArray sixBytes = file.read(6); 
    if (sixBytes.startsWith("solid ")) 
    {
      QByteArray endLine = "endsolid"; 
 
	  while (!file.atEnd())
	  {
        QByteArray line = file.readLine(); 
        if (line.startsWith("facet ")) 
        { 
			while (!file.atEnd()) 
            { 
			  line = file.readLine(); 
			} 
             
			if (!line.startsWith(endLine)) 
			{ 
		      qDebug("%s\n",line.toStdString().c_str()); 
              return -1; 
		    } 
             
            return 1; 
        } 
        else if (line.startsWith(endLine)) 
		{ 
          return 1; 
		}
	  }
      // reset and check for binary... 
      if (!file.reset()) 
      { 
    	qDebug("\n\tCannot seek to the 0th byte (before the header)"); 
        return -1; 
      }
    }

    if (fileSize < 84)
    {
        // 80-byte header + 4-byte "number of triangles" for a binary file
        qDebug("\n\tThe STL file is not long enough (%u bytes).", uint(fileSize));
        return -1;
    }

    // Header is from bytes 0-79; numTriangleBytes starts at byte offset 80.
    if (!file.seek(80))
    {
        qDebug("\n\tCannot seek to the 80th byte (after the header)");
        return -1;
    }

    // Read the number of triangles, uint32_t (4 bytes), little-endian
    QByteArray nTrianglesBytes = file.read(4);
    if (nTrianglesBytes.size() != 4)
    {
        qDebug("\n\tCannot read the number of triangles (after the header)");
        return -1;
    }            

    uint32_t nTriangles = *((uint32_t*)nTrianglesBytes.data());

    // Verify that file size equals the sum of header + nTriangles value + all triangles
    if (fileSize == (84 + (nTriangles * 50)))
        return 0;

    return -1;
}

bool isSTL(std::string filePath)
{
  bool stl = true;
  std::string my_str = filePath.substr(filePath.size()-3, 3);
  transform(my_str.begin(), my_str.end(), my_str.begin(), ::tolower);
  if (my_str != "stl") { stl = false; }

  return stl;
}