#include "pvOptFuncs.h"
#include "math.h"

#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cstring>
#include <sstream>
#include <string>
#include <vector>

#include <QObject>
#include <QDebug>
#include <QFile>
#include <QDir>
#include <QFileDialog>
#include <QWidget>

//-----------------------------------------------------------------------------

pvOptFuncs::pvOptFuncs() 
{
}

//execute terminal command
void pvOptFuncs::executeSystemCmd(std::string command)
{
  system(command.c_str());
}

// extract doubles from string and put in vector
std::vector<double> pvOptFuncs::extractDoubles(std::string str)
{
  std::stringstream stream(str);
  std::string substr;
  std::vector<double> output;
  output.clear();

  while (stream >> substr)
  {
    if (substr.front() == '(')
    {
      substr.erase(substr.begin());
    }
    if (substr.back() == ')')
    {
      substr.pop_back();
    }
    double val = std::atof(substr.c_str());
    if (val != 0)
    {
      output.push_back(val);
    }
  }

  return output;
}

std::string pvOptFuncs::removeExtensionFromFile(std::string fileName)
{
  size_t lastdot = fileName.find_last_of(".");
  if (lastdot == std::string::npos) return fileName;
  return fileName.substr(0, lastdot);
}

std::vector<std::string> pvOptFuncs::splitString(std::string str, std::string delimiter)
{
  std::vector<std::string> stringList;
  int start = 0;
  int end = str.find(delimiter);
  while (end != -1)
  {
    stringList.push_back(str.substr(start, end - start));
    start = end + delimiter.size();
    end = str.find(delimiter, start);
  }
  stringList.push_back(str.substr(start, end - start));

  return stringList;
}

long long pvOptFuncs::calculateGCD(long long a, long long b)
{
  if (a == 0)
    return b;
  else if (b == 0)
    return a;
  if (a < b)
    return calculateGCD(a, b % a);
  else
    return calculateGCD(b, a % b);
}

fraction pvOptFuncs::decimalToFraction(double number)
{
    fraction f;
    int intVal = floor(number);
    double fVal = number - intVal;
    const long pVal = 1000000000;
    long long gcdVal = calculateGCD(round(fVal * pVal), pVal);
    long long num = round(fVal * pVal) / gcdVal;
    long long deno = pVal / gcdVal;
    f.a = (intVal * deno) + num;
    f.b = deno;
    return f;
}

float pvOptFuncs::vectorNorm(std::vector<float> v)
{
  float sum = 0;
  for (int i = 0; i < v.size(); i++)
  {
    sum += pow(v[i],2);
  }
  return sqrt(sum);
}

int pvOptFuncs::roundInt(double r) 
{
  return (r > 0.0) ? (r + 0.5) : (r - 0.5); 
}

// Check docker version, output whether it is up to date
dockerFound pvOptFuncs::checkDocker(std::string rootPath)
{
  std::string dockerDebugFile = rootPath + "/checkDocker.txt";
  std::string check = "docker -v > " + dockerDebugFile;
  executeSystemCmd(check);

  dockerFound s;
  s.success = true;
  s.goodDocker = false;

  QFile file(dockerDebugFile.c_str());
  if (!file.open(QIODevice::ReadOnly))
  {
    s.success = false;
  }
  while (!file.atEnd())
  {
    std::string line = file.readLine().toStdString();
    if (line.find("Docker version") != std::string::npos)
    {
      if (extractDoubles(line)[0]>=19) {  s.goodDocker = true;  }
    }
    s.infoStr = line;
  }
  file.flush();
  file.close();
  file.remove();
  
  return s;
}

// write python gui
std::string pvOptFuncs::writePythonGUI(std::string rootPath)
{
  std::ostringstream pyGuiPath;
  pyGuiPath << rootPath << "/pvGUI.py";
  std::ofstream pyGui(pyGuiPath.str().c_str(), std::ios::out);
  
  pyGui<<
  "#!/usr/bin/env python\n"
  "\n"
  "import os\n"
  "import re\n"
  "import sys\n"
  "import argparse\n"
  "\n"
  "import tkinter as Tk\n"
  "from tkinter import ttk\n"
  "from tkinter import font as tkFont\n"
  "from tkinter import filedialog as fd \n"
  "\n"
  "import matplotlib\n"
  "from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk\n"
  "from matplotlib import pyplot as plt\n"
  "from matplotlib.animation import FuncAnimation\n"
  "import numpy as np\n"
  "\n"
  "parser = argparse.ArgumentParser(description='PvOptGUI Post Processing Interface')\n"
  "parser.add_argument('--animated', action='store_true', help='Option to animate the plot for a running optimization')\n"
  "parser.add_argument('--testing', action='store_true', help='Option to run test, outputs number of iterations and type')\n"
  "parser.add_argument('-log', type=str, nargs='?', help='The log file to read data from')\n"
  "parser.add_argument('-clTitle', type=str, nargs='?', help='The title to display during the CL solution animation')\n"
  "parser.add_argument('-simTitle', type=str, nargs='?', help='The title to display during a single flow solution run')\n"
  "parser.add_argument('-optTitle', type=str, nargs='?', help='The title to display during the optimization animation')\n"
  "args = parser.parse_args()\n"
  "\n"
  "testing = False\n"
  "optimizationRunning = False\n"
  "optLog = ''\n"
  "alphaTitle = ''\n"
  "optTitle = ''\n"
  "simTitle = ''\n"
  "\n"
  "if args.animated != None : optimizationRunning = args.animated\n"
  "if args.testing != None : testing = args.testing \n"
  "if args.log != None : optLog = args.log\n"
  "if args.clTitle != None : alphaTitle = args.clTitle\n"
  "if args.optTitle != None : optTitle = args.optTitle\n"
  "if args.simTitle != None : simTitle = args.simTitle\n"
  "\n"
  "searchAlphaTitle = alphaTitle\n"
  "searchOptTitle = optTitle\n"
  "searchSimTitle = simTitle\n"
  "\n"
  "printInterval = 10\n"
  "alpha = 0\n"
  "numFlow = -1\n"
  "numAdj = -1\n"
  "numObj = int()\n"
  "\n"
  "solveAlphaFound = False\n"
  "alphaFound = False\n"
  "optimizationFound = False\n"
  "solvingFlow = False\n"
  "\n"
  "flowResLabels = ['Ux', 'Uy', 'Uz', 'p', 'he', 'NuTilda', 'k', 'Omega', 'Epsilon']\n"
  "objFuncLabels = ['CD', 'CL', 'CM']\n"
  "colors = ['b','r','g','c','gold','purple','m','y','k','navy','maroon']\n"
  "\n"
  "def getFloats(line):\n"
  "    numeric_const_pattern = '[-+]? (?: (?: \\d* \\. \\d+ ) | (?: \\d+ \\.? ) )(?: [Ee] [+-]? \\d+ ) ?'\n"
  "    rx = re.compile(numeric_const_pattern, re.VERBOSE)\n"
  "    out = rx.findall(line)\n"
  "    return out\n"
  "\n"
  "def listToPlot(vector):\n"
  "    vec = [[float(i) for i in j] for j in vector]\n"
  "    return np.array(list(zip(*(vec))))\n"
  "\n"
  "def getObj(vector):\n"
  "    n = len(vector)\n"
  "    vec = np.zeros([numObj,n])\n"
  "    for i in range(n):\n"
  "        dat = listToPlot(vector[i])\n"
  "        l=10\n"
  "        vec[:,i] = dat[l:l+numObj,-2]\n"
  "    return vec\n"
  "\n"
  "def getLog():\n"
  "    global solveAlphaFound\n"
  "    global alphaFound\n"
  "    global optimizationFound\n"
  "    global optimizationRunning\n"
  "    global solvingFlow\n"
  "\n"
  "    global aFlowVectors\n"
  "    global flowVectors\n"
  "    global adjointVectors\n"
  "    global currentFlowVec\n"
  "    global tempFlowVectors\n"
  "    global tempAdjointVectors\n"
  "    global iterationRelation\n"
  "\n"
  "    global alpha\n"
  "    global numFlow\n"
  "    global numAdj\n"
  "    global numObj\n"
  "    global flowResShown\n"
  "    global flowIt\n"
  "    global adjIt\n"
  "\n"
  "    searchingAdjoint = False\n"
  "    searchingFlow = False\n"
  "    flowIt = -1\n"
  "    sensIt = -1\n"
  "    s = 0\n"
  "    series = -1\n"
  "    alphaSensIt = 0\n"
  "    flowLen = 13\n"
  "    flowResShown = np.zeros(flowLen-4)\n"
  "\n"
  "    #CL Solution vector\n"
  "    aFlowVectors = [[[]]] #flow, residuals vs minor iteration\n"
  "    flowVectors = [[[]]] #flow, residuals vs minor iteration\n"
  "    adjointVectors = [[[]]] #sens, residual vs minor iteration\n"
  "    tempFlowVectors = [[]]\n"
  "    tempAdjointVectors = [[]]\n"
  "    iterationRelation = [['']] #flow iterations, corresponding adjIt\n"
  "    r=0\n"
  "    \n"
  "    if (os.path.exists(optLog)):\n"
  "        logFile = open(optLog, 'r')\n"
  "        lines = logFile.readlines()\n"
  "        for currentLine in lines:\n"
  "            if 'Running SolveCL to find alpha that matches target CL' in currentLine:\n"
  "                alphaFound = True\n"
  "                alpha = 1\n"
  "            elif 'Completed! alpha' in currentLine:\n"
  "                alpha = 0\n"
  "                flowVectors.pop(0)\n"
  "                alphaSensIt = len(flowVectors)\n"
  "                aFlowVectors = flowVectors[:][:][:]\n"
  "                flowVectors = [[[]]]\n"
  "                solveAlphaFound = True\n"
  "            elif 'Optimization Problem -- opt' in currentLine:\n"
  "                optimizationFound = True\n"
  "                r+=1\n"
  "                iterationRelation.append([])\n"
  "            elif 'Evaluating Objective Functions' in currentLine:\n"
  "                flowIt = (int(getFloats(currentLine)[0]) - alphaSensIt if solveAlphaFound else int(getFloats(currentLine)[0]))\n"
  "                searchingFlow = True\n"
  "                searchingAdjoint = False\n"
  "                s = -1\n"
  "                tempFlowVectors = [[0]*flowLen]\n"
  "                iterationRelation[r].append(flowIt)\n"
  "            elif 'Evaluating Objective Function Sensitivities' in currentLine:\n"
  "                adjIt = int(getFloats(currentLine)[0])\n"
  "                searchingAdjoint = True\n"
  "                searchingFlow = False\n"
  "                series = -1   \n"
  "                iterationRelation[r].append(adjIt)\n"
  "                r+=1\n"
  "                iterationRelation.append([])\n"
  "            elif searchingFlow:\n"
  "                residualData = getFloats(currentLine)\n"
  "                if 'U Initial residual:' in currentLine:\n"
  "                    tempFlowVectors.append([0]*flowLen)\n"
  "                    s+=1\n"
  "                    tempFlowVectors[s][0] = (1 if s==0 else s * printInterval)\n"
  "                    tempFlowVectors[s][1] = residualData[0]\n"
  "                    tempFlowVectors[s][2] = residualData[1]\n"
  "                    tempFlowVectors[s][3] = residualData[2]\n"
  "                    flowResShown[:3] = [1,1,1]\n"
  "                elif 'p Initial residual:' in currentLine:\n"
  "                    tempFlowVectors[s][4] = residualData[0]\n"
  "                    flowResShown[3] = 1 \n"
  "                elif 'he Initial residual:' in currentLine:\n"
  "                    tempFlowVectors[s][5] = residualData[0]\n"
  "                    flowResShown[4] = 1\n"
  "                elif 'nuTilda Initial residual:' in currentLine:\n"
  "                    tempFlowVectors[s][6] = residualData[0]\n"
  "                    flowResShown[5] = 1\n"
  "                elif 'k Initial residual:' in currentLine:\n"
  "                    tempFlowVectors[s][7] = residualData[0]\n"
  "                    flowResShown[6] = 1\n"
  "                elif 'omega Initial residual:' in currentLine:\n"
  "                    tempFlowVectors[s][8] = residualData[0]\n"
  "                    flowResShown[7] = 1\n"
  "                elif 'epsilon Initial residual:' in currentLine:\n"
  "                    tempFlowVectors[s][9] = residualData[0]\n"
  "                    flowResShown[8] = 1\n"
  "                elif 'CD-part1-force:' in currentLine:\n"
  "                    tempFlowVectors[s][10] = residualData[1]\n"
  "                elif 'CL-part1-force:' in currentLine:\n"
  "                    tempFlowVectors[s][11] = residualData[1]\n"
  "                elif 'CM-part1-force:' in currentLine:\n"
  "                    tempFlowVectors[s][12] = residualData[1]\n"
  "                elif 'Flow Runtime:' in currentLine:\n"
  "                    flowVectors.append(tempFlowVectors[:][:])\n"
  "\n"
  "            elif searchingAdjoint:\n"
  "                adjointData = getFloats(currentLine)\n"
  "                if 'Main iteration' in currentLine:\n"
  "                    if 'Main iteration 0' in currentLine: \n"
  "                        series += 1; \n"
  "                        t=0\n"
  "                        if (series == 0):\n"
  "                            tempAdjointVectors = [[0]*2*numObj]\n"
  "                    if (series == 0):\n"
  "                        if t != 0 : tempAdjointVectors.append([0]*2*numObj)\n"
  "                        tempAdjointVectors[t][0] = adjointData[0]\n"
  "                        tempAdjointVectors[t][1] = adjointData[1]\n"
  "                    elif (series > 0):\n"
  "                        if t>=len(tempAdjointVectors): tempAdjointVectors.append([0]*2*numObj)\n"
  "                        tempAdjointVectors[t][2*series] = adjointData[0]\n"
  "                        tempAdjointVectors[t][2*series+1] = adjointData[1]\n"
  "                    t+=1\n"
  "                elif 'Adjoint Runtime:' in currentLine : \n"
  "                    adjointVectors.append(tempAdjointVectors[:][:])\n"
  "\n"
  "            if 'Solution:' in currentLine:\n"
  "                if optimizationRunning != 't':\n"
  "                    optimizationRunning = 0\n"
  "\n"
  "        logFile.close()\n"
  "    if flowVectors[0] == [[]] : flowVectors.pop(0)\n"
  "    if adjointVectors[0] == [[]] : adjointVectors.pop(0)\n"
  "    currentFlowVec = aFlowVectors if alpha else flowVectors\n"
  "    for i in range(2) : \n"
  "        if len(iterationRelation) > 0 : iterationRelation.pop(-1)\n"
  "\n"
  "    if not optimizationFound and not alphaFound:\n"
  "        solvingFlow = True\n"
  "    else:\n"
  "        solvingFlow = False\n"
  "\n"
  "def findLogInfo():\n"
  "    global numObj\n"
  "    global searchAlphaTitle\n"
  "    global searchOptTitle\n"
  "    global searchSimTitle\n"
  "    global printInterval\n"
  "\n"
  "    searchingAdjoint = False\n"
  "    searchingFlowBC = False\n"
  "    searchingNormalizedFlow = False\n"
  "    searchingClAlpha = False\n"
  "    searchingIC = False\n"
  "    pIFound = False\n"
  "    machNum = 0\n"
  "    reynoldsNum=0\n"
  "    targetCL = 0.5\n"
  "    meshCells = 0\n"
  "    solver = ''\n"
  "    turbulenceModel = ''\n"
  "    solveClAlpha = 0\n"
  "\n"
  "    if (os.path.exists(optLog)):\n"
  "        logFile = open(optLog, 'r')\n"
  "        lines = logFile.readlines()\n"
  "        o=0\n"
  "        s=0\n"
  "        t=0\n"
  "        e=0\n"
  "        for currentLine in lines:\n"
  "            readData = np.array([float(i) for i in getFloats(currentLine)])\n"
  "            if 'Selecting RAS turbulence model' in currentLine and s==0:\n"
  "                turbulenceModel = currentLine.split()[4]\n"
  "                s=1\n"
  "            elif 'Global Cells:' in currentLine:\n"
  "                meshCells = int(readData[0])\n"
  "            elif 'solverName' in currentLine:\n"
  "                solver = currentLine.split()[1].rstrip(';')\n"
  "            elif 'primalBC' in currentLine:\n"
  "                searchingFlowBC = True\n"
  "                flowData = np.zeros([4])\n"
  "                findFlow = np.zeros([5])\n"
  "            elif 'normalizeStates' in currentLine:\n"
  "                searchingFlowBC = False\n"
  "                searchingNormalizedFlow = True\n"
  "            elif 'objFunc' in currentLine and o==0:\n"
  "                searchingNormalizedFlow = False\n"
  "                searchingAdjoint = True\n"
  "                machNum = flowData[0] / np.sqrt(flowData[1]*1.4*287.0)\n"
  "                reynoldsNum = int(flowData[0]*flowData[3]/flowData[2])\n"
  "            elif 'Evaluating Objective Functions' in currentLine and e==0:\n"
  "                searchingAdjoint = False\n"
  "                searchingClAlpha = True\n"
  "                e=1\n"
  "            elif 'printInterval' in currentLine and pIFound==False:\n"
  "                printInterval = int(readData[0])\n"
  "                pIFound = True\n"
  "            elif 'Objective Function: calcObjFuncValues' in currentLine:\n"
  "                searchingIC = True\n"
  "                t+=1\n"
  "            elif searchingFlowBC:\n"
  "                if 'U0' in currentLine:\n"
  "                    findFlow = np.array([1,0,0])\n"
  "                elif 'T0' in currentLine:\n"
  "                    findFlow = np.array([0,1,0])\n"
  "                elif 'nuTilda0' in currentLine:\n"
  "                    findFlow = np.array([0,0,1])\n"
  "                if findFlow[0]==1 and ('value' in currentLine):\n"
  "                    flowData[0] = readData[1]\n"
  "                    findFlow = np.array([0,0,0])\n"
  "                elif findFlow[1]==1 and ('value' in currentLine):\n"
  "                    flowData[1] = readData[1]\n"
  "                    findFlow = np.array([0,0,0])\n"
  "                elif findFlow[2]==1 and ('value' in currentLine):\n"
  "                    r=10\n"
  "                    flowData[2] = readData[1]/r\n"
  "                    findFlow = np.array([0,0,0])\n"
  "            elif searchingNormalizedFlow:\n"
  "                if 'p_rgh' in currentLine:\n"
  "                    flowData[3] = readData[0]\n"
  "            elif searchingAdjoint:\n"
  "                if 'part1' in currentLine: \n"
  "                    o+=1\n"
  "                elif 'designVar' in currentLine:\n"
  "                    numObj = o\n"
  "            elif searchingClAlpha:\n"
  "                if 'alpha' in currentLine:\n"
  "                    solveClAlpha = readData[0]\n"
  "                    searchingClAlpha = False\n"
  "            elif searchingIC:\n"
  "                if 'alpha_0' in currentLine:\n"
  "                    if t==1 : optimizationAlpha = readData[3]\n"
  "                    #if t==2 : finalAlpha = readData[3]\n"
  "                if 'CL' in currentLine:\n"
  "                    targetCL = readData[1]\n"
  "                    searchingIC = False\n"
  "        logFile.close()\n"
  "\n"
  "        if not solvingFlow:\n"
  "            searchOptTitle = 'Optimization\\nMach Number: ' + str(round(machNum,2)) + ', Reynolds Number: ' + str(reynoldsNum) + ', Target CL: ' + str(targetCL) + '\\nMesh Cells: ' + str(meshCells) + ', Solver: ' + str(solver) + ', Turbulence Model: ' + str(turbulenceModel)\n"
  "            searchAlphaTitle = 'CL Solution\\nAlpha0: ' + str(solveClAlpha) + ', Target CL: ' + str(targetCL)\n"
  "        else:\n"
  "            searchSimTitle = 'Solving Flow\\nAlpha: ' + str(solveClAlpha) + ', Mach Number: ' + str(round(machNum,2)) + ', Reynolds Number: ' + str(reynoldsNum) + '\\nMesh Cells: ' + str(meshCells) + ', Solver: ' + str(solver) + ', Turbulence Model: ' + str(turbulenceModel)\n"
  "\n"
  "def quit():\n"
  "    plt.close()\n"
  "    root.quit()\n"
  "    root.destroy()\n"
  "    sys.exit()\n"
  "\n"
  "def change_historyType(event):\n"
  "    global alpha\n"
  "    global numFlow\n"
  "    global numAdj\n"
  "    numFlow = -1\n"
  "    numAdj = -1\n"
  "    histStr = tkHistory.get()\n"
  "    if histStr == 'CL Solution':\n"
  "        alpha = 1\n"
  "    else:\n"
  "        alpha = 0\n"
  "    redrawPlot([numFlow, numAdj])\n"
  "\n"
  "def change_flowIteration(event):\n"
  "    global numFlow\n"
  "    global numAdj\n"
  "    global tkFlowIt\n"
  "    numFlow = int(-1 if tkFlowIt.get() == '' else tkFlowIt.get())\n"
  "    if alpha == 0 and isinstance(numFlow,int):\n"
  "        ex = 0\n"
  "        for i in iterationRelation[1:]:\n"
  "            if (numFlow in i[:-1]) and (ex == 0) and isinstance(i[0],int):\n"
  "                ex = 1\n"
  "                numAdj = i[-1]\n"
  "    update_labels()\n"
  "    redrawPlot([numFlow, numAdj])\n"
  "\n"
  "def change_adjIteration(event):\n"
  "    global numAdj\n"
  "    global numFlow\n"
  "    global tkAdjIt\n"
  "    numAdj = int(-1 if tkAdjIt.get() == '' else tkAdjIt.get())\n"
  "    if alpha == 0 and isinstance(numAdj,int):\n"
  "        ex = 0\n"
  "        for i in iterationRelation[1:]:\n"
  "            if i[-1] == numAdj and ex == 0 and isinstance(i[0],int):\n"
  "                ex = 1\n"
  "                numFlow = i[-2]\n"
  "    update_labels()\n"
  "    redrawPlot([numFlow, numAdj])\n"
  "\n"
  "def open_log():\n"
  "    global optLog\n"
  "    global alpha\n"
  "    global alphaFound\n"
  "    global numFlow\n"
  "    global numAdj\n"
  "    global optimizationFound\n"
  "    global history_box\n"
  "    global optTitle\n"
  "    global alphaTitle\n"
  "    global simTitle\n"
  "    global optimizationRunning\n"
  "    global solvingFlow\n"
  "\n"
  "    optimizationRunning = 0\n"
  "    alphaFound = False\n"
  "    optimizationFound = False\n"
  "    solvingFlow = False\n"
  "    optLog = fd.askopenfilename()\n"
  "    getLog()\n"
  "    findLogInfo()\n"
  "    optTitle = searchOptTitle\n"
  "    alphaTitle = searchAlphaTitle\n"
  "    simTitle = searchSimTitle\n"
  "    alpha = 1 if alphaFound and not optimizationFound else 0\n"
  "    numFlow = -1\n"
  "    numAdj = -1\n"
  "    update_labels()\n"
  "    history_box.current(len(history_box['values'])-1)\n"
  "    redrawPlot([numFlow, numAdj])\n"
  "\n"
  "def update_labels():\n"
  "    global adj_iteration\n"
  "    global flow_iteration\n"
  "    global currentFlowVec\n"
  "    global history_box\n"
  "    global numObj\n"
  "\n"
  "    if alpha == 1 or len(adj_iteration['values'])==0:\n"
  "        adj_iteration['values'] = ['']\n"
  "        adj_iteration.current(0)\n"
  "    elif alpha == 0 and optimizationFound:\n"
  "        adj_iteration['values'] = [i for i in range(len(adjointVectors))]\n"
  "        adj_iteration.current(range(len(adjointVectors))[numAdj])\n"
  "    currentFlowVec = aFlowVectors if alpha else flowVectors\n"
  "\n"
  "    if len(flow_iteration['values'])==0:\n"
  "        flow_iteration['values'] = ['']\n"
  "        flow_iteration.current(0)\n"
  "    else:\n"
  "        flow_iteration['values'] = [i for i in range(len(currentFlowVec))]\n"
  "        flow_iteration.current(range(len(currentFlowVec))[numFlow])\n"
  "        \n"
  "    historyType = []\n"
  "    if alphaFound: historyType.append('CL Solution')\n"
  "    if optimizationFound: historyType.append('Optimization')\n"
  "    if len(history_box['values'])==0: historyType.append('Flow Solution')\n"
  "    history_box['values'] = historyType\n"
  "\n"
  "    objVectors = getObj(currentFlowVec)\n"
  "    if len(objVectors[0,:])>0:\n"
  "        for i in range(numObj):\n"
  "            objFuncValue_labels[i].config(text=str('%.8f'%(objVectors[i,numFlow])))\n"
  "    optObjVectors = getObj(flowVectors)\n"
  "    if len(optObjVectors[0,:])>0:\n"
  "        for i in range(numObj):\n"
  "            baselineObjFuncValue_labels[i].config(text=str('%.8f'%(optObjVectors[i,0])))\n"
  "\n"
  "\n"
  "def initAnimation():\n"
  "    global fig\n"
  "    global figureInfo\n"
  "    global patches\n"
  "    global legends\n"
  "    global flowResShown\n"
  "    global plotLines\n"
  "\n"
  "    subfigs = figureInfo[0]\n"
  "    subFigsNest = figureInfo[1]\n"
  "    flowAxs = figureInfo[2]\n"
  "    adjAxs = figureInfo[3]\n"
  "    itObjAxs = figureInfo[4]\n"
  "    totObjAxs = figureInfo[5]\n"
  "    for line in flowAxs.lines : line.remove()\n"
  "    for line in adjAxs.lines : line.remove()\n"
  "    for i in itObjAxs : \n"
  "        for line in i.lines : line.remove()\n"
  "    for i in totObjAxs : \n"
  "        for line in i.lines : line.remove()\n"
  "\n"
  "    #Initialize plot lines\n"
  "    flowLines = [flowAxs.semilogy([],[],label=flowResLabels[i],color=colors[i])[0] for i in range(len(flowResLabels))]\n"
  "    adjLines = [adjAxs.semilogy([],[],label=objFuncLabels[i],color=colors[i])[0] for i in range(len(objFuncLabels))]\n"
  "    itObjLines = [itObjAxs[numObj-(i+1)].plot([],[],label=objFuncLabels[i],color=colors[i])[0] for i in range(len(objFuncLabels))]\n"
  "    totObjLines = [totObjAxs[numObj-(i+1)].plot([],[],label=objFuncLabels[i],color=colors[i])[0] for i in range(len(objFuncLabels))]\n"
  "    plotLines = np.array([flowLines,adjLines,itObjLines,totObjLines], dtype=list)\n"
  "    patches = np.sum(plotLines)\n"
  "\n"
  "    #Initialize Labels\n"
  "    flowAxs.set_xlabel('Flow Iterations')\n"
  "    flowAxs.set_ylabel('Flow Residuals')\n"
  "    flowAxs.grid(True)\n"
  "    adjAxs.set_xlabel('Adjoint Iterations')\n"
  "    adjAxs.set_ylabel('Adjoint Residuals')\n"
  "    adjAxs.grid(True)\n"
  "    itObjAxs[numObj-1].set_xlabel('Flow Iterations')\n"
  "    s=numObj-1\n"
  "    for ax in itObjAxs:\n"
  "        ax.set_ylabel(objFuncLabels[s])\n"
  "        ax.grid(True)\n"
  "        s-=1\n"
  "    subFigsNest[1].suptitle('Optimization History')\n"
  "    totObjAxs[numObj-1].set_xlabel('Flow Solution #')\n"
  "    s=numObj-1\n"
  "    for ax in totObjAxs:\n"
  "        ax.set_ylabel(objFuncLabels[s])\n"
  "        ax.grid(True)\n"
  "        s-=1\n"
  "\n"
  "    patches = np.sum(plotLines)\n"
  "    return patches\n"
  "\n"
  "def animatePlot(frame):\n"
  "    global tempFlowVectors\n"
  "    global tempAdjointVectors\n"
  "    global adjointVectors\n"
  "    global currentFlowVec\n"
  "    global fig\n"
  "    global patches\n"
  "    global plotLines\n"
  "    global ani\n"
  "\n"
  "    getLog()\n"
  "    fig.suptitle(alphaTitle if alpha else (simTitle if solvingFlow else optTitle), fontsize=int(root.winfo_height()/ppi+1))\n"
  "    legends = [[] for i in range(4)]\n"
  "    flowData = np.array([])\n"
  "\n"
  "    lns = []\n"
  "    rolling=[[],[]]\n"
  "    flowLines = plotLines[0]\n"
  "    flowAxs.set_title('Flow Solution # ' + str(flowIt), fontsize='large')\n"
  "    if len(tempFlowVectors) > 1:\n"
  "        flowData = listToPlot(tempFlowVectors)\n"
  "        for i in range(len(flowResLabels)):\n"
  "            if flowResShown[i]==1 and len(flowData[i+1,:])>1:\n"
  "                r = np.where(flowData[i+1,:]==0.0)[0][0] if np.any(flowData[i+1]==0.0) else None\n"
  "                flowLines[i].set_data(flowData[0,:r],flowData[i+1,:r])\n"
  "                lns += [flowLines[i]]\n"
  "                rolling[0] += [np.min(flowData[i+1,:r])]\n"
  "                rolling[1] += [np.max(flowData[i+1,:r])]\n"
  "        legends[0] = flowAxs.legend(lns, [l.get_label() for l in lns],loc='upper right')\n"
  "        lim=np.max(flowData[0,:])\n"
  "        flowAxs.set_xlim([-0.05*lim,lim*1.05])\n"
  "        flowAxs.set_ylim([0.8*np.min(rolling[0]),1.2*np.max(rolling[1])])\n"
  "\n"
  "    lns = []\n"
  "    rolling=[[],[]]\n"
  "    adjLines = plotLines[1]\n"
  "    adjAxs.set_title('Adjoint Solution # ' + ('' if (alpha or len(adjointVectors) == 0) else str(adjIt)), fontsize='large')\n"
  "    if len(tempAdjointVectors) > 1 and optimizationFound:\n"
  "        adjData = listToPlot(tempAdjointVectors)\n"
  "        for i in range(numObj):\n"
  "            r = np.where(adjData[2*i+1,:]==0.0)[0][0] if np.any(adjData[2*i+1,1:]==0.0) else -1\n"
  "            adjLines[i].set_data(adjData[2*i,:r], adjData[2*i+1,:r])\n"
  "            lns += [adjLines[i]]\n"
  "            if np.size(adjData[2*i+1,:r]) != 0:\n"
  "                rolling[0] += [np.min(adjData[2*i+1,:r])]\n"
  "                rolling[1] += [np.max(adjData[2*i+1,:r])]\n"
  "        legends[1] = adjAxs.legend(lns.__reversed__(), [l.get_label() for l in lns].__reversed__(),loc='upper right')\n"
  "        lim = max(adjData[0,:])\n"
  "        adjAxs.set_xlim([-0.05*lim,lim*1.05])\n"
  "        adjAxs.set_ylim([0.8*np.min(rolling[0]),1.2*np.max(rolling[1])])\n"
  "\n"
  "    lns = []\n"
  "    itObjLines = plotLines[2]\n"
  "    subFigsNest[0].suptitle('Flow Solution # ' + str(flowIt))\n"
  "    for i in range(numObj): \n"
  "        if len(tempFlowVectors) > 1:\n"
  "            r = np.where(flowData[i+10,:]==0)[0][0] if np.any(flowData[i+10]==0.0) else -1\n"
  "            itObjLines[i].set_data(flowData[0,:r], flowData[i+10,:r])\n"
  "            lns += [itObjLines[i]]\n"
  "            lim=max(flowData[0,:])\n"
  "            itObjAxs[(numObj-1)-i].set_xlim([-0.05*lim,lim*1.05])\n"
  "            itObjAxs[(numObj-1)-i].set_ylim([0.97*np.min(flowData[i+10,:r]) if np.min(flowData[i+10,:r]) > 0 else 1.03 * np.min(flowData[i+10,:r]),1.03*np.max(flowData[i+10,:r]) if np.max(flowData[i+10,:r])>0 else 0.97*np.max(flowData[i+10,:r])])\n"
  "    legends[2] = itObjAxs[0].legend(lns.__reversed__(), [l.get_label() for l in lns].__reversed__(),loc='upper right')\n"
  "    \n"
  "    lns = []\n"
  "    totObjLines = plotLines[3]\n"
  "\n"
  "    if len(flowVectors) > 0:\n"
  "        objVectors = getObj(flowVectors)\n"
  "        for i in range(numObj): \n"
  "            objMarker = ' ' if len(objVectors[i,:]) > 1 else '.'\n"
  "            totObjLines[i].set_marker(objMarker)\n"
  "            totObjLines[i].set_data(range(len(objVectors[i,:])), objVectors[i,:])\n"
  "            lns += [totObjLines[i]]\n"
  "            lim = len(objVectors[i])\n"
  "            totObjAxs[(numObj-1)-i].set_xlim([-0.05*lim,lim*1.05])\n"
  "            totObjAxs[(numObj-1)-i].set_ylim([0.98*np.min(objVectors[i]) if np.min(objVectors[i])>0 else 1.02*np.min(objVectors[i]),1.02*np.max(objVectors[i]) if np.max(objVectors[i])>0 else 0.98*np.max(objVectors[i])])\n"
  "        legends[3] = totObjAxs[0].legend(lns.__reversed__(), [l.get_label() for l in lns].__reversed__(),loc='upper right')\n"
  "    \n"
  "        if len(objVectors[0,:])>0:\n"
  "            for i in range(numObj):\n"
  "                objFuncValue_labels[i].config(text=str('%.8f'%(objVectors[i,-1])))\n"
  "        optObjVectors = getObj(flowVectors)\n"
  "        if len(optObjVectors[0,:])>0:\n"
  "            for i in range(numObj):\n"
  "                baselineObjFuncValue_labels[i].config(text=str('%.8f'%(optObjVectors[i,0])))\n"
  "\n"
  "    patches = flowLines+adjLines+itObjLines+totObjLines+legends\n"
  "    if optimizationRunning==0: \n"
  "        ani._stop()\n"
  "        findLogInfo()\n"
  "        redrawPlot([-1,-1])\n"
  "    return patches\n"
  "\n"
  "\n"
  "def redrawPlot(n):\n"
  "    global fig\n"
  "    global figureInfo\n"
  "    global canvas\n"
  "    global subfigs\n"
  "    global subFigsNest\n"
  "    global flowAxs\n"
  "    global adjAxs\n"
  "    global itObjAxs\n"
  "    global totObjAxs\n"
  "    global currentFlowVec\n"
  "    global adjointVectors\n"
  "\n"
  "    currentFlowVec = aFlowVectors if alpha else flowVectors\n"
  "\n"
  "    fig.suptitle(alphaTitle if alpha else (simTitle if solvingFlow else optTitle), fontsize=int(screen_height/ppi+1))\n"
  "\n"
  "    #Initialize Labels\n"
  "    flowAxs.clear()\n"
  "    flowAxs.set_title('Flow Solution # ' + ('' if len(currentFlowVec) == 0 else str(currentFlowVec.index(currentFlowVec[n[0]]))), fontsize='large')\n"
  "    flowAxs.set_xlabel('Flow Iterations')\n"
  "    flowAxs.set_ylabel('Flow Residuals')\n"
  "    flowAxs.grid(True)\n"
  "    adjAxs.clear()\n"
  "    adjAxs.set_title('Adjoint Solution # ' + ('' if (alpha or len(adjointVectors) == 0) else str(adjointVectors.index(adjointVectors[n[1]]))), fontsize='large')\n"
  "    adjAxs.set_xlabel('Adjoint Iterations')\n"
  "    adjAxs.set_ylabel('Adjoint Residuals')\n"
  "    adjAxs.grid(True)\n"
  "    for i in itObjAxs: i.clear()\n"
  "    subFigsNest[0].suptitle('Flow Solution # ' + ('' if len(currentFlowVec) == 0  else str(currentFlowVec.index(currentFlowVec[n[0]]))))\n"
  "    itObjAxs[-1].set_xlabel('Flow Iterations')\n"
  "    s=numObj-1\n"
  "    for ax in itObjAxs:\n"
  "        ax.set_ylabel(objFuncLabels[s])\n"
  "        ax.grid(True)\n"
  "        s-=1\n"
  "    for i in totObjAxs: i.clear()\n"
  "    subFigsNest[1].suptitle('Optimization History')\n"
  "    totObjAxs[-1].set_xlabel('Flow Solution #')\n"
  "    s=numObj-1\n"
  "    for ax in totObjAxs:\n"
  "        ax.set_ylabel(objFuncLabels[s])\n"
  "        ax.grid(True)\n"
  "        s-=1\n"
  "\n"
  "    #plot\n"
  "    lns=[]\n"
  "    if len(currentFlowVec) > 0:\n"
  "        if len(currentFlowVec[n[0]]) > 0:\n"
  "            flowData = listToPlot(currentFlowVec[n[0]])\n"
  "            for i in range(len(flowResLabels)):\n"
  "                if np.all(flowData[i+1,1:-1] != 0) :\n"
  "                    r = np.where(flowData[i+1,:]==0)[0][0]\n"
  "                    lns+=[flowAxs.semilogy(flowData[0,:r],flowData[i+1,:r],label=flowResLabels[i],color=colors[i])[0]]\n"
  "            flowAxs.legend(lns, [l.get_label() for l in lns])\n"
  "\n"
  "    lns=[]\n"
  "    if len(adjointVectors)>0:\n"
  "        if len(adjointVectors[numAdj]) > 1 and optimizationFound and not alpha:\n"
  "            adjData = listToPlot(adjointVectors[numAdj])\n"
  "            for i in range(numObj):\n"
  "                r = np.where(adjData[2*i+1,:] == 0)[0][0] if np.any(adjData[2*i+1,:]==0.0) else -1\n"
  "                lns+=[adjAxs.semilogy(adjData[2*i,:r], adjData[2*i+1,:r],label=objFuncLabels[i],color=colors[i])[0]]\n"
  "            adjAxs.legend(lns.__reversed__(), [l.get_label() for l in lns].__reversed__())\n"
  "\n"
  "    lns=[[],[]]\n"
  "    if len(currentFlowVec) > 0:\n"
  "        objVectors = getObj(currentFlowVec)\n"
  "        for i in range(numObj): \n"
  "            if np.all(flowData[i+10,1:-1] != 0):\n"
  "                r = np.where(flowData[i+10,:]==0)[0][0]\n"
  "                lns[0]+=[itObjAxs[(numObj-1)-i].plot(flowData[0,:r], flowData[i+10,:r], label=objFuncLabels[i], color=colors[i])[0]]\n"
  "            r = n[0]+1 if n[0]!=-1 else None\n"
  "            objMarker = ' ' if len(objVectors[i,:r]) > 1 else '.'\n"
  "            lns[1]+=[totObjAxs[(numObj-1)-i].plot(range(len(objVectors[i,:r])), objVectors[i,:r], label=objFuncLabels[i], color=colors[i], marker=objMarker)[0]]\n"
  "        itObjAxs[-1].legend(lns[0].__reversed__(), [l.get_label() for l in lns[0]].__reversed__())\n"
  "        totObjAxs[-1].legend(lns[1].__reversed__(), [l.get_label() for l in lns[1]].__reversed__())\n"
  "\n"
  "    update_labels()\n"
  "    flow_iteration['values'] = [i for i in range(len(currentFlowVec))]\n"
  "    adj_iteration['values'] = [i for i in range(len(adjointVectors))]\n"
  "    canvas.draw()\n"
  "    canvas.flush_events()\n"
  "\n"
  "\n"
  "###Run###\n"
  "\n"
  "while(numObj == 0):\n"
  "    findLogInfo()\n"
  "    if numObj==0 and optimizationRunning!=1:\n"
  "        numObj = 2\n"
  "getLog()\n"
  "if not testing:\n"
  "    matplotlib.use('TkAgg')\n"
  "    root = Tk.Tk()\n"
  "    tkHistory = Tk.StringVar(root)\n"
  "    tkFlowIt = Tk.StringVar(root)\n"
  "    tkAdjIt = Tk.StringVar(root)\n"
  "\n"
  "    screen_width = root.winfo_screenwidth()\n"
  "    screen_height = root.winfo_screenheight()\n"
  "    ppi=96\n"
  "    fig = plt.figure(dpi=ppi, facecolor='white', constrained_layout=True)\n"
  "    #canvas = Tk.Canvas(master=root, fig)\n"
  "    canvas = FigureCanvasTkAgg(fig, master=root)\n"
  "    canvas.draw_idle()\n"
  "    #Initialize subfigures and plot locations\n"
  "    subfigs = fig.subfigures(nrows=2,ncols=1,wspace=2.0)\n"
  "    (flowAxs, adjAxs) = subfigs[0].subplots(1,2)\n"
  "    #flowAxs.autoscale_on(True)\n"
  "    subFigsNest = subfigs[1].subfigures(1,2, hspace=None,wspace=None)\n"
  "    itObjAxs = subFigsNest[0].subplots(numObj, sharex=True)\n"
  "    totObjAxs = subFigsNest[1].subplots(numObj, sharex=True)\n"
  "    figureInfo = [subfigs,subFigsNest,flowAxs,adjAxs,itObjAxs,totObjAxs]\n"
  "\n"
  "    ### tk interface\n"
  "    root.wm_title('pvOptAirfoil')\n"
  "    root.protocol('WM_DELETE_WINDOW', quit)\n"
  "    root.geometry('960x640')\n"
  "    cols = 4\n"
  "    rows = 20\n"
  "    for i in range(rows): root.grid_rowconfigure(i,weight=1)\n"
  "    for i in range(cols): root.grid_columnconfigure(i,weight=1)\n"
  "    canvas.get_tk_widget().grid(column=0,row=0,columnspan=cols-1,rowspan=rows-1,sticky='NSEW')\n"
  "    font = tkFont.Font(family='Helvetica Heavy', size=int(screen_height/ppi))\n"
  "    toolbar = NavigationToolbar2Tk(canvas, root, pack_toolbar=False)\n"
  "    toolbar.grid(column=0,row=rows-1)\n"
  "    toolbar.update()\n"
  "    main_frame = Tk.Frame(root)\n"
  "    main_frame.grid(column=cols-1,row=0,rowspan=rows-2,columnspan=1,sticky='NSEW')\n"
  "\n"
  "    #Simulation\n"
  "    for i in range(12): main_frame.grid_rowconfigure(i, weight=1)\n"
  "    for i in range(2): main_frame.grid_columnconfigure(i, weight=1)\n"
  "    history_label = Tk.Label(main_frame, text='Case History', font=font)\n"
  "    history_label.grid(column=0, row=1, sticky='NSEW')\n"
  "    #Simulaton Combo Box\n"
  "    historyType = []\n"
  "    if alphaFound: historyType.append('CL Solution')\n"
  "    if optimizationFound: historyType.append('Optimization')\n"
  "    if len(historyType)==0: historyType.append('Flow Solution')\n"
  "    history_box = ttk.Combobox(main_frame, textvariable=tkHistory, values=historyType, font=font)\n"
  "    if len(historyType)>0 : history_box.current(len(history_box['values'])-1)\n"
  "    history_box.bind('<<ComboboxSelected>>', change_historyType)\n"
  "    history_box.grid(column=1, row=1, sticky='nesw')\n"
  "\n"
  "    #Flow Iteration Title\n"
  "    flow_iteration_label = Tk.Label(main_frame, text='Flow Solution #', font=font)\n"
  "    flow_iteration_label.grid(column=0, row=2, sticky='nesw')\n"
  "    #Flow Iteration Combo Box\n"
  "    flow_iteration = ttk.Combobox(main_frame, textvariable=tkFlowIt, font=font)\n"
  "    if len(currentFlowVec)>0: \n"
  "        flow_iteration['values'] = [i for i in range(len(currentFlowVec))]\n"
  "        #flow_iteration.current(currentFlowVec.index(currentFlowVec[numFlow]))\n"
  "    flow_iteration.bind('<<ComboboxSelected>>',change_flowIteration)\n"
  "    flow_iteration.grid(column=1, row=2, sticky='nesw')\n"
  "\n"
  "    #Adjoint Iteration Title\n"
  "    adj_iteration_label = Tk.Label(main_frame, text='Adjoint Solution #', font=font)\n"
  "    adj_iteration_label.grid(column=0, row=3, sticky='nesw')\n"
  "    #Adjoint Iteration Combo Box\n"
  "    adj_iteration = ttk.Combobox(main_frame, textvariable=tkAdjIt, font=font)\n"
  "    if len(adjointVectors)>0:\n"
  "        adj_iteration['values'] = [i for i in range(len(adjointVectors))]\n"
  "        #adj_iteration.current(len(adj_iteration['values'])-1)\n"
  "    adj_iteration.bind('<<ComboboxSelected>>',change_adjIteration)\n"
  "    adj_iteration.grid(column=1, row=3, sticky='nesw')\n"
  "\n"
  "    #Choose logFile\n"
  "    file_label = Tk.Label(main_frame, text='Select Different Log', font=font)\n"
  "    file_label.grid(column=0, row=4, sticky='nesw')\n"
  "    #file select button\n"
  "    file_button = Tk.Button(main_frame, text='Open File', command=open_log, font=font)\n"
  "    file_button.grid(column=1, row=4, sticky='nesw')\n"
  "\n"
  "    #Latest Obj Func Values\n"
  "    #frame config\n"
  "    objFuncRows = 2*numObj+4\n"
  "    objFunc_frame = Tk.Frame(main_frame, borderwidth = 3, relief='ridge')\n"
  "    objFunc_frame.grid(column=0,row=6,rowspan=objFuncRows,columnspan=2,sticky='nesw')\n"
  "    for i in range(objFuncRows): objFunc_frame.grid_rowconfigure(i, weight=1)\n"
  "    for i in range(2): objFunc_frame.grid_columnconfigure(i, weight=1)\n"
  "    # values\n"
  "    baselineObjFunc_label = Tk.Label(objFunc_frame, text='Baseline Function Values', font=font)\n"
  "    baselineObjFunc_label.grid(column=0, columnspan=2, row=0, sticky='nesw')\n"
  "    baselineObjFunc_labels = [i for i in range(numObj)]\n"
  "    baselineObjFuncValue_labels = [i for i in range(numObj)]\n"
  "    for i in range(numObj):\n"
  "        baselineObjFunc_labels[i] = Tk.Label(objFunc_frame, text=objFuncLabels[i]+':', font=font)\n"
  "        baselineObjFunc_labels[i].grid(column=0, row=i+1, sticky='nesw')\n"
  "\n"
  "        baselineObjFuncValue_labels[i] = Tk.Label(objFunc_frame, text='', font=font, borderwidth = 1, relief='sunken')\n"
  "        baselineObjFuncValue_labels[i].grid(column=1, row=i+1, sticky='nesw')\n"
  "\n"
  "    currentObjFunc_label = Tk.Label(objFunc_frame, text='Current Function Values', font=font)\n"
  "    currentObjFunc_label.grid(column=0, columnspan=2, row=2+numObj, sticky='nesw')\n"
  "    objFunc_labels = [i for i in range(numObj)]\n"
  "    objFuncValue_labels = [i for i in range(numObj)]\n"
  "    for i in range(numObj):\n"
  "        objFunc_labels[i] = Tk.Label(objFunc_frame, text=objFuncLabels[i]+':', font=font)\n"
  "        objFunc_labels[i].grid(column=0, row=i+3+numObj, sticky='nesw')\n"
  "\n"
  "        objFuncValue_labels[i] = Tk.Label(objFunc_frame, text='', font=font, borderwidth = 1, relief='sunken')\n"
  "        objFuncValue_labels[i].grid(column=1, row=i+3+numObj, sticky='nesw')\n"
  "\n"
  "    \n"
  "    #init\n"
  "    useBlit = False #Blitted will not update titles or axis ticks/labels, only lines and other axes artists can be updated. Blit gives big speed boost\n"
  "    if optimizationRunning==1:\n"
  "        ani = FuncAnimation(fig,func=animatePlot,frames=1,init_func=initAnimation,blit=useBlit,repeat_delay=25)\n"
  "        ani.repeat = True\n"
  "        ani._start()\n"
  "    else:\n"
  "        optTitle = searchOptTitle\n"
  "        alphaTitle = searchAlphaTitle\n"
  "        simTitle = searchSimTitle\n"
  "        root.after(1,redrawPlot([-1,-1]))\n"
  "    root.mainloop()\n"
  "else:\n"
  "    print(str(len(flowVectors))+' flow iterations')\n"
  "    print(str(len(adjointVectors))+' adj iterations')" << std::endl;

  pyGui.flush();
  pyGui.close();

  return pyGuiPath.str();
}

//link case file
paths pvOptFuncs::linkCaseFile(QWidget *parent)
{
  paths p;
  // open file browser
  p.caseFileDir = QFileDialog::getExistingDirectory(parent, QObject::tr("Select Optimization Folder"), QDir::homePath(), QFileDialog::ShowDirsOnly | QFileDialog::DontResolveSymlinks);
  p.rootPath = p.caseFileDir.toStdString();
  return p;
}
